# -*- coding: utf-8 -*-
"""fd-vpn.modules.vpn_client"""

import logging
import re
import time
import uuid

from uuid import uuid4
from datetime import datetime, timedelta

import bcrypt
import pyotp

from six import ensure_str, iteritems

from fd_vpn.classes.modules import VPNModuleClientBase
from fd_vpn.objects.vpn_client import FdVpnObjectVPNClient
from fd_vpn.objects.vpn_server import FdVpnObjectVPNServer
from fd_vpn.objects.vpn_group import FdVpnObjectVPNGroup

from dwho.classes.modules import DWhoModuleSQLBase, MODULES
from sonicprobe import helpers
from sonicprobe.libs import xys
from sonicprobe.libs.moresynchro import RWLock
from sonicprobe.validator.email import SPValidatorEmail
from httpdis.ext.httpdis_json import HttpReqError, HttpReqErrJson


LOG = logging.getLogger('fd-vpn.modules.vpn_client')

xys.add_list('vpn_client.auth_method', FdVpnObjectVPNClient.get_auth_method_list())
xys.add_list('vpn_client.sort_by', FdVpnObjectVPNClient.get_columns() + ['vpn_group.groupname'])
xys.add_list('vpn_client.order_by', ('asc', 'desc'))

xys.add_regex('vpn_client.reference', re.compile(r'^[a-zA-Z0-9]{8}$').match)
xys.add_callback('vpn_client.email_address', SPValidatorEmail.validate)
xys.add_callback('vpn_client.password_check', helpers.password_check)


class VPNClientModule(DWhoModuleSQLBase, VPNModuleClientBase):
    MODULE_NAME = 'vpn_client'

    LOCK        = RWLock()

    def __init__(self):
        self.objects    = {'vpn_client': FdVpnObjectVPNClient(),
                           'vpn_server': FdVpnObjectVPNServer(),
                           'vpn_group':  FdVpnObjectVPNGroup()}

        DWhoModuleSQLBase.__init__(self)
        VPNModuleClientBase.__init__(self)

    def init(self, config):
        DWhoModuleSQLBase.init(self, config)
        VPNModuleClientBase.init(self, config)

        self.objects['vpn_client'].init(config)
        self.objects['vpn_server'].init(config)
        self.objects['vpn_group'].init(config)


    INDEX_SCHEMA    = xys.load("""
    offset?:        !~~digit
    limit?:         !~~uint
    filter?:        !!any
    sort_by?:       !~~isIn(vpn_client.sort_by)
    sort_order?:    !~~isIn(vpn_client.order_by)
    """)

    INDEX_DEFAULTS  = {'offset':        0,
                       'limit':         10,
                       'filter':        {},
                       'sort_by':       'vpn_client.id',
                       'sort_order':    'asc'}

    def index(self, request):
        options = request.query_params()
        if not xys.validate(options, self.INDEX_SCHEMA):
            raise HttpReqErrJson(415, "invalid arguments for command")

        for key, value in iteritems(self.INDEX_DEFAULTS):
            options.setdefault(key, value)

        if not self.LOCK.acquire_read(self.lock_timeout):
            raise HttpReqErrJson(503, "unable to take LOCK for reading after %s seconds" % self.lock_timeout)

        r       = {'results':   [],
                   'total':     0}

        try:
            cnt = self.objects['vpn_client'].get_count()
            res = self.objects['vpn_client'].get_all_server_group(to_dict = True,
                                                                  order   = ((options['sort_by'], options['sort_order']),),
                                                                  limit   = int(options['limit']),
                                                                  offset  = int(options['offset']))
            if res:
                for row in res:
                    row['vpn_client.ip_addr'] = '.'.join(row['vpn_server.cidr'].split('.')[0:3] + [row['vpn_client.last_octet']])
                    row.pop('vpn_client.auth_data', None)

                r['results']    = res
                r['total']      = cnt

            return r
        except HttpReqErrJson:
            raise
        except Exception as e:
            LOG.exception(e)
            raise HttpReqErrJson(503, repr(e))
        finally:
            self.LOCK.release()


    FREE_LAST_OCTETS_SCHEMA  = xys.load("""
    server_id: !~~digit
    client_id: !~~digit
    """)

    def free_last_octets(self, request):
        self.plugin_credentials.set_request(request)
        if not self.plugin_credentials.is_admin():
            raise HttpReqErrJson(403, "bad credentials")

        options = request.query_params()
        if not xys.validate(options, self.FREE_LAST_OCTETS_SCHEMA):
            raise HttpReqErrJson(415, "invalid arguments for command")

        if not self.LOCK.acquire_read(self.lock_timeout):
            raise HttpReqErrJson(503, "unable to take LOCK for reading after %s seconds" % self.lock_timeout)
        try:
            server      = self.objects['vpn_server'].get(**{'id': options['server_id']})
            if not server:
                raise HttpReqErrJson(404, "unable to find server. (id: %r)" % options['server_id'])

            server_ip   = '.'.join(server['cidr'].split('.')[0:3])
            last_octets = self.objects['vpn_client'].get_free_last_octets(server['id'], options['client_id'])

            if not last_octets:
                raise HttpReqErrJson(507, "no more free ip address available")

            return [{'last_octet':  last_octet,
                     'addr_ip':     "%s.%s" % (server_ip, last_octet)} for last_octet in last_octets]
        except HttpReqErrJson:
            raise
        except Exception as e:
            LOG.exception(e)
            raise HttpReqErrJson(503, repr(e))
        finally:
            self.LOCK.release()


    FORM_SCHEMA  = xys.load("""
    id?:    !~~digit
    action: !~~enum(create,update)
    """)

    def form(self, request):
        self.plugin_credentials.set_request(request)
        if not self.plugin_credentials.is_admin():
            raise HttpReqError(403, "bad credentials")

        options = request.query_params()
        if not xys.validate(options, self.FORM_SCHEMA):
            raise HttpReqError(415, "invalid arguments for command")

        servers = self.objects['vpn_server'].get_all_group(to_dict    = True,
                                                           order      = 'vpn_server.id',
                                                           with_group = True)

        if options['action'] == 'create':
            return self.render('form.html',
                               action       = 'create',
                               xid          = '',
                               auth_methods = FdVpnObjectVPNClient.get_auth_method_list(),
                               vpn_client   = {},
                               servers      = servers or [])

        if not self.LOCK.acquire_read(self.lock_timeout):
            raise HttpReqError(503, "unable to take LOCK for reading after %s seconds" % self.lock_timeout)

        try:
            xobject             = self._fetch(options['id'])
            if not xobject:
                raise HttpReqError(404, "client not found. (id: %r)" % options['id'])

            server              = self.objects['vpn_server'].get(**{'id': xobject['server_id']})
            if not server:
                raise HttpReqError(404, "server not found. (id: %r)" % xobject['server_id'])

            xobject['ip_addr']  = '.'.join(server['cidr'].split('.')[0:3] + [xobject['last_octet']])

            return self.render('form.html',
                               action       = 'update',
                               xid          = options['id'],
                               auth_methods = FdVpnObjectVPNClient.get_auth_method_list(),
                               vpn_client   = xobject,
                               servers      = servers or [])
        except HttpReqError:
            raise
        except Exception as e:
            LOG.exception(e)
            raise HttpReqError(503, repr(e))
        finally:
            self.LOCK.release()


    GET_SCHEMA  = xys.load("""
    id: !~~digit
    """)

    def get(self, request):
        self.plugin_credentials.set_request(request)
        if not self.plugin_credentials.is_admin():
            raise HttpReqErrJson(403, "bad credentials")

        options = request.query_params()
        if not xys.validate(options, self.GET_SCHEMA):
            raise HttpReqErrJson(415, "invalid arguments for command")

        client = self._fetch(options['id'], secure = False)
        if not client:
            raise HttpReqErrJson(404, "client not found. (id: %r)" % options['id'])

        server = self.objects['vpn_server'].get(**{'id': client['server_id']})
        if not server:
            raise HttpReqErrJson(404, "server not found. (id: %r)" % client['server_id'])

        return self._get(server, client)


    DOWNLOAD_SCHEMA  = xys.load("""
    id:   !~~digit
    arch: !~~enum(x32,x64,xp_x32,xp_x64,ovpn)
    """)

    def download(self, request):
        self.plugin_credentials.set_request(request)
        if not self.plugin_credentials.is_admin():
            raise HttpReqErrJson(403, "bad credentials")

        options = request.query_params()
        if not xys.validate(options, self.DOWNLOAD_SCHEMA):
            raise HttpReqErrJson(415, "invalid arguments for command")

        client  = self._fetch(options['id'])
        if not client:
            raise HttpReqErrJson(404, "unable to find client. (id: %r)" % options['id'])

        server  = self.objects['vpn_server'].get(to_dict = True,
                                                 **{'id': client['server_id']})
        if not server:
            raise HttpReqErrJson(404, "unable to find server. (id: %r)" % client['server_id'])

        return self._download(server, client, options['arch'])


    CREATE_SCHEMA   = xys.load("""
    server_id:         !~~digit
    last_octet:        !~~digit
    auth_method?:      !~~isIn(vpn_client.auth_method)
    auth_data?-:       !!str
    comment*[0,1000]-: !!str
    """)

    def create(self, request):
        self.plugin_credentials.set_request(request)
        if not self.plugin_credentials.is_admin():
            raise HttpReqErrJson(403, "bad credentials")

        args = request.payload_params()
        if not xys.validate(args, self.CREATE_SCHEMA):
            raise HttpReqErrJson(415, "invalid arguments for command")

        if not self.LOCK.acquire_write(self.lock_timeout):
            raise HttpReqErrJson(503, "unable to take LOCK for writing after %s seconds" % self.lock_timeout)

        try:
            server              = self.objects['vpn_server'].get(**{'id': args['server_id']})
            if not server:
                raise HttpReqErrJson(404, "unable to find server. (id: %r)" % args['server_id'])

            if not server['group_id']:
                raise HttpReqErrJson(415, "invalid server. (id: %r)" % server['id'])

            last_octets         = self.objects['vpn_client'].get_free_last_octets(server['id'])
            if not last_octets:
                raise HttpReqErrJson(507, "no more free ip address available")

            if int(args['last_octet']) not in last_octets:
                raise HttpReqErrJson(415, "invalid address ip. (last_octet: %r)" % args['last_octet'])

            data                = dict(args)
            data['id']          = None
            data['reference']   = str(uuid.uuid4())[0:8]
            data['enabled']     = True

            if data['auth_method'] == 'otp':
                data['auth_data'] = pyotp.random_base32()
            elif data['auth_method'] == 'userpass':
                try:
                    helpers.password_check(data['auth_data'])
                except Exception as e:
                    raise HttpReqErrJson(415, repr(e))
                data['auth_data'] = bcrypt.hashpw(ensure_str(data['auth_data']),
                                                  bcrypt.gensalt())
            else:
                data['auth_data'] = None

            xobject             = self._fetch(
                                    self.objects['vpn_client'].create(
                                        data.keys(),
                                        data.values()))

            if not int(server['enabled']) or not server['group_id']:
                self.plugin_iptables.add_drop_src('SERVER', server['cidr'])
                self.plugin_iptables.add_drop_dst('SERVER', server['cidr'])
            else:
                self.plugin_iptables.delete_drop_src('SERVER', server['cidr'])
                self.plugin_iptables.delete_drop_dst('SERVER', server['cidr'])

            self.plugin_openvpn.create_server_config(
                self.plugin_openvpn.build_server_config(server))

            self.plugin_openvpn.create_ccd_config(server['reference'],
                                                  server['cidr'],
                                                  xobject)

            self.plugin_openvpn.cmd_call('enable', server['reference'])
            #self.plugin_openvpn.cmd_call('restart', server['reference'])
            return xobject
        except HttpReqErrJson:
            raise
        except Exception as e:
            LOG.exception(e)
            raise HttpReqErrJson(503, repr(e))
        finally:
            self.LOCK.release()

    def _update(self, xid, data):
        return self.objects['vpn_client'].update(data.keys(),
                                                 data.values(),
                                                 **{'id': xid})


    UPDATE_SCHEMA   = xys.load("""
    id:                !~~digit
    group_id?:         !~~digit
    server_id:         !~~digit
    last_octet:        !~~digit
    auth_method?:      !~~isIn(vpn_client.auth_method)
    auth_data?-:       !!str
    comment*[0,1000]-: !!str
    groupname?:        !!str
    ip_addr?:          !!str
    """)

    def update(self, request):
        self.plugin_credentials.set_request(request)
        if not self.plugin_credentials.is_admin():
            raise HttpReqErrJson(403, "bad credentials")

        args = request.payload_params()
        if not xys.validate(args, self.UPDATE_SCHEMA):
            raise HttpReqErrJson(415, "invalid arguments for command")

        if not self.LOCK.acquire_write(self.lock_timeout):
            raise HttpReqErrJson(503, "unable to take LOCK for writing after %s seconds" % self.lock_timeout)

        try:
            server      = self.objects['vpn_server'].get(**{'id': args['server_id']})
            if not server:
                raise HttpReqErrJson(404, "unable to find server. (id: %r)" % args['server_id'])

            if not server['group_id']:
                raise HttpReqErrJson(415, "invalid server. (id: %r)" % server['id'])

            xid         = args.pop('id')
            xdict       = {'server_id':     args['server_id'],
                           'last_octet':    args['last_octet'],
                           'auth_method':   args['auth_method'],
                           'comment':       args['comment']}

            xobject     = self._fetch(xid, secure = False)
            if not xobject:
                raise HttpReqErrJson(404, "unable to find client. (id: %r)" % xid)

            if xdict['auth_method'] == 'otp':
                if xobject['auth_method'] != xdict['auth_method']:
                    xdict['auth_data'] = pyotp.random_base32()
                elif not xobject['auth_data']:
                    xdict['auth_data'] = pyotp.random_base32()
            elif xdict['auth_method'] == 'userpass':
                if xobject['auth_method'] != xdict['auth_method'] \
                   and not args.get('auth_data'):
                    raise HttpReqErrJson(415, "missing password")

                if args.get('auth_data'):
                    try:
                        helpers.password_check(args['auth_data'])
                    except Exception as e:
                        raise HttpReqErrJson(415, repr(e))
                    xdict['auth_data'] = bcrypt.hashpw(ensure_str(args['auth_data']),
                                                       bcrypt.gensalt())
            else:
                xdict['auth_data'] = None

            last_octets = self.objects['vpn_client'].get_free_last_octets(server['id'], xobject['id'])
            if not last_octets:
                raise HttpReqErrJson(507, "no more free ip address available")

            if int(args['last_octet']) not in last_octets \
               and int(args['last_octet']) != int(xobject['last_octet']):
                raise HttpReqErrJson(415, "invalid address ip. (last_octet: %r)" % args['last_octet'])

            if xobject['server_id']:
                server_old      = self.objects['vpn_server'].get(**{'id': xobject['server_id']})
                if server_old:
                    self.plugin_openvpn.remove_ccd_config(server_old['reference'],
                                                          xobject)

            if not self._update(xid, xdict):
                raise HttpReqErrJson(500, "unable to update client. (id: %r)" % xid)

            xobject     = self._fetch(xid)

            if not int(server['enabled']) or not server['group_id']:
                self.plugin_iptables.add_drop_src('SERVER', server['cidr'])
                self.plugin_iptables.add_drop_dst('SERVER', server['cidr'])
            else:
                self.plugin_iptables.delete_drop_src('SERVER', server['cidr'])
                self.plugin_iptables.delete_drop_dst('SERVER', server['cidr'])

            self.plugin_openvpn.create_server_config(
                self.plugin_openvpn.build_server_config(server))

            self.plugin_openvpn.cmd_call('enable', server['reference'])
            #self.plugin_openvpn.cmd_call('restart', server['reference'])

            self.plugin_openvpn.create_ccd_config(server['reference'],
                                                  server['cidr'],
                                                  xobject)
            return xobject
        except HttpReqErrJson:
            raise
        except Exception as e:
            LOG.exception(e)
            raise HttpReqErrJson(503, repr(e))
        finally:
            self.LOCK.release()


    _TOGGLE_ENABLED_SCHEMA  = xys.load("""
    id: !~~digit
    """)

    def _toggle_enabled(self, request, enabled):
        self.plugin_credentials.set_request(request)
        if not self.plugin_credentials.is_admin():
            raise HttpReqErrJson(403, "bad credentials")

        options = request.query_params()
        if not xys.validate(options, self._TOGGLE_ENABLED_SCHEMA):
            raise HttpReqErrJson(415, "invalid arguments for command")

        if not self.LOCK.acquire_write(self.lock_timeout):
            raise HttpReqErrJson(503, "unable to take LOCK for writing after %s seconds" % self.lock_timeout)

        try:
            xid = options.pop('id')

            if not self._update(xid, {'enabled': helpers.boolize(enabled)}):
                raise HttpReqErrJson(500, "unable to update client. (id: %r)" % xid)

            xobject = self._fetch(xid)
            if not xobject:
                raise HttpReqErrJson(404, "unable to find client. (id: %r)" % xid)

            server  = self.objects['vpn_server'].get(**{'id': xobject['server_id']})
            if not server:
                raise HttpReqErrJson(404, "unable to find server. (id: %r)" % xobject['server_id'])

            addr_ip = '.'.join(server['cidr'].split('.')[0:3] + [xobject['last_octet']])

            if not int(xobject['enabled']):
                self.plugin_iptables.add_drop_src('CLIENT', addr_ip)
                self.plugin_iptables.add_drop_dst('CLIENT', addr_ip)
                self.plugin_openvpn.disable_ccd_config(server['reference'], xobject)
            else:
                self.plugin_iptables.delete_drop_src('CLIENT', addr_ip)
                self.plugin_iptables.delete_drop_dst('CLIENT', addr_ip)
                self.plugin_openvpn.enable_ccd_config(server['reference'], xobject)

            return xobject
        except HttpReqErrJson:
            raise
        except Exception as e:
            LOG.exception(e)
            raise HttpReqErrJson(503, repr(e))
        finally:
            self.LOCK.release()

    def enable(self, request):
        return self._toggle_enabled(request, True)

    def disable(self, request):
        return self._toggle_enabled(request, False)


    DELETE_SCHEMA  = xys.load("""
    id: !~~digit
    """)

    def delete(self, request):
        self.plugin_credentials.set_request(request)
        if not self.plugin_credentials.is_admin():
            raise HttpReqErrJson(403, "bad credentials")

        options = request.query_params()
        if not xys.validate(options, self.DELETE_SCHEMA):
            raise HttpReqErrJson(415, "invalid arguments for command")

        if not self.LOCK.acquire_write(self.lock_timeout):
            raise HttpReqErrJson(503, "unable to take LOCK for writing after %s seconds" % self.lock_timeout)

        try:
            xid = options.pop('id')

            xobject = self._fetch(xid)
            if not xobject:
                raise HttpReqErrJson(404, "unable to find client. (id: %r)" % xid)

            server  = self.objects['vpn_server'].get(**{'id': xobject['server_id']})
            if not server:
                raise HttpReqErrJson(404, "unable to find server. (id: %r)" % xobject['server_id'])

            self.plugin_openvpn.remove_ccd_config(server['reference'], xobject)
            self.objects['vpn_client'].delete(**{'id': xid})
        except HttpReqErrJson:
            raise
        except Exception as e:
            LOG.exception(e)
            raise HttpReqErrJson(503, repr(e))
        finally:
            self.LOCK.release()


    _TOGGLE_CONNECTED_SCHEMA  = xys.load("""
    reference: !~~regex(vpn_client.reference)
    """)

    def _toggle_connected(self, request, connected):
        options = request.query_params()
        if not xys.validate(options, self._TOGGLE_CONNECTED_SCHEMA):
            raise HttpReqErrJson(415, "invalid arguments for command")

        if not self.LOCK.acquire_write(self.lock_timeout):
            raise HttpReqErrJson(503, "unable to take LOCK for writing after %s seconds" % self.lock_timeout)

        try:
            reference = options.pop('reference')

            xdict = {'connected': helpers.boolize(connected)}

            if connected:
                xdict['connected_at'] = time.strftime('%Y-%m-%d %H:%M:%S')
            else:
                xdict['disconnected_at'] = time.strftime('%Y-%m-%d %H:%M:%S')

            xobject = self._fetch_by_reference(reference)
            if not xobject:
                raise HttpReqErrJson(404, "unable to find client. (reference: %r)" % reference)

            if not xobject['enabled']:
                raise HttpReqErrJson(404, "client disabled. (reference: %r)" % reference)

            self._update(xobject['id'], xdict)

            server  = self.objects['vpn_server'].get(**{'id': xobject['server_id']})
            if not server:
                raise HttpReqErrJson(404, "unable to find server. (id: %r)" % xobject['server_id'])

            if not server['enabled']:
                raise HttpReqErrJson(404, "server disabled. (reference: %r)" % server['reference'])

            return True
        except HttpReqErrJson:
            raise
        except Exception as e:
            LOG.exception(e)
            raise HttpReqErrJson(503, repr(e))
        finally:
            self.LOCK.release()

    def connect(self, request):
        return self._toggle_connected(request, True)

    def disconnect(self, request):
        return self._toggle_connected(request, False)


    AUTHENTICATE_QSCHEMA  = xys.load("""
    reference: !~~regex(vpn_client.reference)
    """)

    AUTHENTICATE_PSCHEMA  = xys.load("""
    password-: !!str
    """)

    def authenticate(self, request):
        options = request.query_params()
        if not xys.validate(options, self.AUTHENTICATE_QSCHEMA):
            raise HttpReqErrJson(415, "invalid arguments for command")

        args = request.payload_params()
        if not xys.validate(args, self.AUTHENTICATE_PSCHEMA):
            raise HttpReqErrJson(415, "invalid arguments for command")

        if not self.LOCK.acquire_write(self.lock_timeout):
            raise HttpReqErrJson(503, "unable to take LOCK for writing after %s seconds" % self.lock_timeout)

        try:
            reference = options.pop('reference')

            xobject = self._fetch_by_reference(reference, secure = False)
            if not xobject:
                raise HttpReqErrJson(404, "unable to find client. (reference: %r)" % reference)

            if not xobject['enabled']:
                raise HttpReqErrJson(404, "client disabled. (reference: %r)" % reference)

            server  = self.objects['vpn_server'].get(**{'id': xobject['server_id']})
            if not server:
                raise HttpReqErrJson(404, "unable to find server. (id: %r)" % xobject['server_id'])

            if not server['enabled']:
                raise HttpReqErrJson(404, "server disabled. (reference: %r)" % server['reference'])

            if xobject['auth_method'] == 'userpass':
                return bcrypt.checkpw(ensure_str(args['password']),
                                      ensure_str(xobject['auth_data']))

            if xobject['auth_method'] == 'otp':
                return pyotp.totp.TOTP(xobject['auth_data']).verify(ensure_str(args['password']))

            return False
        except HttpReqErrJson:
            raise
        except Exception as e:
            LOG.exception(e)
            raise HttpReqErrJson(503, repr(e))
        finally:
            self.LOCK.release()

    SEND_MAIL_PSCHEMA = xys.load("""
    reference:     !~~regex(vpn_client.reference)
    email_address: !~~callback(vpn_client.email_address)
    """)

    def send_mail(self, request):
        args = request.payload_params()
        if not xys.validate(args, self.SEND_MAIL_PSCHEMA):
            raise HttpReqErrJson(415, "invalid arguments for command")

        if not self.LOCK.acquire_write(self.lock_timeout):
            raise HttpReqErrJson(503, "unable to take LOCK for writing after %s seconds" % self.lock_timeout)

        try:
            reference = args.pop('reference')

            xobject = self._fetch_by_reference(reference, secure = False)
            if not xobject:
                raise HttpReqErrJson(404, "unable to find client. (reference: %r)" % reference)

            if not xobject['enabled']:
                raise HttpReqErrJson(404, "client disabled. (reference: %r)" % reference)

            server  = self.objects['vpn_server'].get(**{'id': xobject['server_id']})
            if not server:
                raise HttpReqErrJson(404, "unable to find server. (id: %r)" % xobject['server_id'])

            if not server['enabled']:
                raise HttpReqErrJson(404, "server disabled. (reference: %r)" % server['reference'])

            xdict = {}
            xdict['security_reference'] = "%s" % uuid4()
            xdict['security_url'] = self.security_link['url'] % xdict
            xdict['security_expired_at'] = datetime.strftime(
                                              datetime.today() + timedelta(seconds = int(self.security_link['expiration'])),
                                              '%Y-%m-%d %H:%M:%S')

            if not self._update(xobject['id'], {'security_reference': xdict['security_reference'],
                                                'security_expired_at': xdict['security_expired_at']}):
                raise HttpReqErrJson(500, "unable to update client. (id: %r)" % xobject['id'])

            return helpers.send_email(self.security_link['from'],
                                      args['email_address'],
                                      self.security_link['subject'] % xdict,
                                      self.security_link['body'] % xdict)
        except HttpReqErrJson:
            raise
        except Exception as e:
            LOG.exception(e)
            raise HttpReqErrJson(503, repr(e))
        finally:
            self.LOCK.release()


    VALIDATE_SCHEMA = xys.load("""
    server_id:         !~~digit
    last_octet:        !~~digit
    auth_method?:      !~~isIn(vpn_client.auth_method)
    comment?[0,1000]-: !!str
    email_address-:    !~~callback(vpn_client.email_address)
    auth_data-:        !~~callback(vpn_client.password_check)
    """)

    def validate(self, request):
        self.plugin_credentials.set_request(request)
        if not self.plugin_credentials.is_admin():
            raise HttpReqErrJson(403, "bad credentials")

        options = request.query_params()
        if not self.LOCK.acquire_read(self.lock_timeout):
            raise HttpReqErrJson(503, "unable to take LOCK for reading after %s seconds" % self.lock_timeout)

        try:
            for key, value in iteritems(options):
                if not self.VALIDATE_SCHEMA.has_key(key):
                    raise HttpReqErrJson(404, "unknown argument for command. (key: %r)" % key)

                if not xys.validate({key: value}, {key: self.VALIDATE_SCHEMA[key]}):
                    raise HttpReqErrJson(415, "invalid argument for command. (key: %r)" % key)

            return True
        except HttpReqErrJson:
            raise
        except Exception as e:
            LOG.exception(e)
            raise HttpReqErrJson(503, repr(e))
        finally:
            self.LOCK.release()


if __name__ != "__main__":
    def _start():
        MODULES.register(VPNClientModule())
    _start()
